"""resource action method decorators specific for the 'byoml' service"""

from functools import wraps

from simple_rest_client.exceptions import ErrorWithResponse

from ._exceptions import (
    ByomlActionError
)


def byoml_exception_decorator(action_method):
    """
    decorator that parses json error responses
    """
    @wraps(action_method)
    def wrapped(*args, **kwargs):
        try:
            return action_method(*args, **kwargs)
        except ErrorWithResponse as exc:
            raise ByomlActionError.from_cause(exc) from exc
    return wrapped



"""exceptions specific to the Byoml Service"""

from ...exceptions import RestResponseError, RestResponseParseError


class ByomlActionError(RestResponseError):
    """error that represents the json messages of a byoml response"""

    @property
    def message(self):
        """the main user error returned by an analytics error response"""
        return self._get_from_body('error', super().message)


class ByomlActionParseError(RestResponseParseError, ByomlActionError):
    """Indicates that a byoml response could not be parsed"""



import argparse
import logging


def main():
    logging.basicConfig()
    parser = argparse.ArgumentParser(
        prog='waylaycli', description='Command line interface to the Waylay Python SDK'
    )
    args = parser.parse_args()
    parser.print_help()




"""REST client for the Waylay Time Series Analytics Service"""

from .._base import WaylayService

from .model import ModelResource
from .about import AboutResource


class ByomlService(WaylayService):
    """REST client for the Waylay BYOML Service"""
    service_key = 'byoml'
    config_key = 'byoml'
    default_root_url = 'https://byoml.waylay.io'
    resource_definitions = {
        'model': ModelResource,
        'about': AboutResource
    }

    model: ModelResource
    about: AboutResource

    def __init__(self):
        super().__init__(
            json_encode_body=True
        )



"""REST definitions for the 'model' entity of the 'byoml' service"""
from zipfile import ZipFile
from io import BytesIO
from functools import wraps
from typing import Callable, List
import joblib

import pandas as pd
import numpy as np

from .._base import WaylayResource
from .._decorators import (
    return_body_decorator,
    return_path_decorator,
    suppress_header_decorator
)
from ._decorators import (
    byoml_exception_decorator
)
from ...exceptions import (
    RestRequestError
)


def _input_data_as_list(input_data):
    if isinstance(input_data, list):
        return input_data
    if isinstance(input_data, pd.DataFrame):
        return input_data.values.tolist()
    # numpy nd_array
    try:
        return input_data.tolist()
    except (ValueError, TypeError):
        pass
    raise RestRequestError(
        f'input data of unsupported type {type(input_data)}'
    )


def model_execution_request_decorator(action_method):
    """
    Decorator that prepares an execution of the model (predict, regress, classify, ...)
    """
    @wraps(action_method)
    def wrapped(model_name, input_data, **kwargs):
        request_body = {
            'instances': _input_data_as_list(input_data)
        }
        return action_method(
            model_name,
            body=request_body,
            **kwargs
        )
    return wrapped


DEFAULT_DECORATORS = [byoml_exception_decorator, return_body_decorator]


def _execute_model_decorators(response_key: str) -> List[Callable]:
    return [
        byoml_exception_decorator,
        model_execution_request_decorator,
        return_path_decorator(
            [response_key],
            default_response_constructor=np.array
        )
    ]


DEFAULT_BYOML_MODEL_TIMEOUT = 60


class ModelResource(WaylayResource):
    """REST Resource for the 'model' entity of the 'byoml' service"""
    actions = {
        'list': {
            'method': 'GET',
            'url': '/models',
            'decorators': [
                byoml_exception_decorator,
                return_path_decorator(['available_models'])
            ],
            'description': 'List the metadata of the deployed <em>BYOML Models</em>'},
        'list_names': {
            'method': 'GET',
            'url': '/models',
            'decorators': [
                byoml_exception_decorator,
                return_path_decorator(['available_models', 'name'])
            ],
            'description': 'List the names of deployed <em>BYOML Models</em>'},
        '_create': {
            'method': 'POST',
            'url': '/models',
            'decorators': DEFAULT_DECORATORS,
            'description': 'Build and create a new <em>BYOML Model</em> as specified in the request'},
        '_replace': {
            'method': 'PUT',
            'url': '/models/{}',
            'decorators': DEFAULT_DECORATORS,
            'description': 'Build and replace the named <em>BYOML Model</em>'},
        'get': {
            'method': 'GET',
            'url': '/models/{}',
            'decorators': DEFAULT_DECORATORS,
            'description': 'Fetch the metadata of the named <em>BYOML Model</em>'},
        'examples': {
            'method': 'GET',
            'url': '/models/{}/examples',
            'decorators': [
                byoml_exception_decorator,
                return_path_decorator(['example_payloads'])
            ],
            'description': 'Fetch the <em>example request input</em> of the named <em>BYOML Model</em>'},
        'predict': {
            'method': 'POST',
            'url': '/models/{}/predict',
            'decorators': _execute_model_decorators('predictions'),
            'description': 'Execute the <em>predict</em> capability of the named <em>BYOML Model</em>'},
        'regress': {
            'method': 'POST',
            'url': '/models/{}/regress',
            'decorators': _execute_model_decorators('result'),
            'description': 'Execute the <em>regress</em> capability of the named  <em>BYOML Model</em>'},
        'classify': {
            'method': 'POST',
            'url': '/models/{}/classify',
            'decorators': _execute_model_decorators('result'),
            'description': 'Execute the <em>classification</em> capability of the named <em>BYOML Model</em>'},
        'remove': {
            'method': 'DELETE',
            'url': '/models/{}',
            'decorators': DEFAULT_DECORATORS,
            'description': 'Remove the named <em>BYOML Model</em>'},
    }

    def __init__(self, *args, **kwargs):
        kwargs.pop('timeout', None)
        super().__init__(*args, timeout=DEFAULT_BYOML_MODEL_TIMEOUT, **kwargs)

    def _send_model_arguments(
        self, model_name, trained_model,
        framework="sklearn", description=""
    ):
        """uploads a binary model with given name, framework and description."""
        # write the model to a file, and put the file in a zipped buffer
        joblib.dump(trained_model, './model.joblib')
        model_zip_buffer = BytesIO()
        with ZipFile(model_zip_buffer, 'w') as zipper:
            zipper.write('model.joblib')
        return {
            'body': {
                "name": model_name,
                "framework": framework,
                "description": description
            },
            'files': {
                "file": ('model.zip', model_zip_buffer.getvalue())
            },
        }

    @suppress_header_decorator('Content-Type')
    def upload(
        self, model_name, trained_model,
        framework="sklearn", description="", **kwargs
    ):
        """uploads a new machine learning model with given name, framework and description."""
        return getattr(self, '_create')(
            **self._send_model_arguments(
                model_name, trained_model,
                framework=framework, description=description,
            ),
            **kwargs
        )

    @suppress_header_decorator('Content-Type')
    def replace(
        self, model_name, trained_model,
        framework="sklearn", description="", **kwargs
    ):
        """replaces a machine learning model with given name, framework and description."""
        return getattr(self, '_replace')(
            model_name,
            **self._send_model_arguments(
                model_name, trained_model,
                framework=framework, description=description,
            ),
            **kwargs
        )



"""Base classes for Waylay REST Services and Resources"""

from typing import (
    Optional, Type, TypeVar,
    Mapping
)
try:
    from typing import Protocol
except ImportError:
    # typing.Protocol is a 3.8 feature ...
    # ... but typing_extensions provides forward compatibility.
    from typing_extensions import Protocol  # type: ignore

from simple_rest_client.api import Resource, API
from waylay.config import WaylayConfig

S = TypeVar('S', bound='WaylayService')
R = TypeVar('R', bound='WaylayResource')


class WaylayServiceContext(Protocol):
    """view protocol for the dynamic service context"""

    def get(self, service_class: Type[S]) -> Optional[S]:
        """returns the service instances for the given class, if available"""

    def require(self, service_class: Type[S]) -> S:
        """returns the service instances for the given class, raise ConfigError otherwise"""


DEFAULT_SEVICE_TIMEOUT = 10


class WaylayResource(Resource):
    """
    Client object representing a Waylay Resource, i.e. a collection of REST
    operations that have a single Waylay Entity as subject.
    """

    def add_action(self, action_name: str):
        """add action, and expect whether any decorators need to be applied"""
        super().add_action(action_name)
        self.decorate_action(action_name)

    def decorate_action(self, action_name):
        """decorates the action if a 'decorators' definition exist"""
        action = self.get_action(action_name)
        decorators = action.get('decorators', None)
        if decorators:
            for decorator in decorators:
                setattr(self, action_name, decorator(getattr(self, action_name)))

    def __repr__(self):
        actions_repr = ', '.join(
            f"{name}: '{action_def['method']} {action_def['url']}'"
            for name, action_def in self.actions.items()
        )
        return (
            f"<{self.__class__.__name__}("
            f"actions=[{actions_repr}]"
            ")>"
        )


class WaylayService(API):
    """
    Client object representing a Waylay Service, i.e. a collection of
    Resources with their operations.
    """
    service_key: str = ''
    config_key: str = 'api'
    default_root_url: str = ''
    config: Optional[WaylayConfig] = None
    resource_definitions: Mapping[str, Type[Resource]] = {}
    plugin_priority = 0

    def __init__(self, *args, **kwargs):
        timeout = kwargs.pop('timeout', DEFAULT_SEVICE_TIMEOUT)
        super().__init__(*args, timeout=timeout, **kwargs)
        for name, resource_class in self.resource_definitions.items():
            self.add_resource(resource_name=name, resource_class=resource_class)

    def _add_waylay_resource(self, resource_name: str, resource_class: Type[R], **kwargs) -> R:
        """add resource and return the initialized result"""
        self.add_resource(resource_name=resource_name, resource_class=resource_class, **kwargs)
        waylay_resource: R = self._resources[self.correct_attribute_name(resource_name)]
        return waylay_resource

    def set_root_url(self, root_url):
        self.config.set_root_url(self.config_key, root_url)
        self.reconfigure()

    def get_root_url(self) -> Optional[str]:
        if self.config is None:
            return self.default_root_url
        return self.config.get_root_url(self.config_key, self.default_root_url)

    def configure(self: S, config: WaylayConfig, context: WaylayServiceContext) -> S:
        """configure endpoints and authentication with given config. Returns self"""
        self.config = config
        return self.reconfigure()

    def reconfigure(self: S) -> S:
        if self.config is None:
            return self
        root_url = self.get_root_url()
        for resource in self._resources.values():
            resource.api_root_url = root_url
            resource.client.auth = self.config.auth
        return self

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}("
            f"service_key={self.service_key},"
            f"config_key={self.config_key},"
            f"root_url={self.get_root_url()},"
            f"resources=[{', '.join(self._resources.keys())}]"
            ")>"
        )



"""Waylay Rest Services"""
from .analytics import AnalyticsService
from ._base import WaylayService, WaylayResource, WaylayServiceContext
from .byoml import ByomlService
from . import _decorators as decorators

SERVICES = [
    AnalyticsService, ByomlService
]



"""Base exception class hierarchy for errors in the waylay client"""
from simple_rest_client.exceptions import ErrorWithResponse as _ErrorWithResponse


class WaylayError(Exception):
    """Root class for all exceptions raised by this module"""


class AuthError(WaylayError):
    """Exception class for waylay authentication errors"""


class ConfigError(WaylayError):
    """Exception class for waylay client configuration"""


class RestError(WaylayError):
    """Exception class for failures to make a rest call"""


class RestRequestError(RestError):
    """
    Exception class for failures to prepare a rest call
    (before http execution)
    """


class RestResponseError(RestError):
    """Exception class wrapping the response data of a Rest call"""

    def __init__(self, message, response):
        super().__init__(message)
        self.response = response

    @property
    def message(self):
        """basic user message for this error"""
        return self.args[0]

    def __str__(self):
        return (
            f"{self.__class__.__name__}({self.response.status_code}: " +
            f"'{self.message}'; {self.response.method} '{self.response.url}')"
        )

    def _get_from_body(self, key, default_value):
        error_resp_body = self.response.body
        if isinstance(error_resp_body, dict):
            return error_resp_body.get(key, default_value)
        return default_value

    @classmethod
    def from_cause(cls, cause: _ErrorWithResponse):
        """converts a rest client error to a waylay client error"""
        return cls(cause.message.split(',')[0], cause.response)


class RestResponseParseError(RestResponseError):
    """Exception raised when a successfull http request (200) could not be parsed succesfully."""



"""definitions for the 'query' entity of the 'analytics' Service"""

from .._base import WaylayResource
from .._decorators import (
    return_path_decorator
)
from ._decorators import (
    analytics_exception_decorator,
    analytics_return_dataframe_decorator,
    MultiFrameHandling
)


def parse_query_entity(json_dict: dict):
    """default response constructor for query config entities"""
    return json_dict


CONFIG_ENTITY_DECORATORS = [
    analytics_exception_decorator,
    return_path_decorator(['query'], parse_query_entity)
]

CONFIG_LIST_DECORATORS = [
    analytics_exception_decorator,
    return_path_decorator(['queries', 'name'])
]

DATA_RESPONSE_DECORATORS = [
    analytics_exception_decorator,
    analytics_return_dataframe_decorator(
        'data',
        default_frames_handling=MultiFrameHandling.JOIN
    )
]


class QueryResource(WaylayResource):
    """REST Resource for the 'query' entity of the 'analytics' Service"""
    actions = {
        'list': {
            'method': 'GET',
            'url': '/config/query',
            'decorators': CONFIG_LIST_DECORATORS,
            'description': 'List or search <em>Query Configurations</em>'},
        'create': {
            'method': 'POST',
            'url': '/config/query',
            'decorators': CONFIG_ENTITY_DECORATORS,
            'description': 'Create a new <em>Query Configuration</em>'},
        'get': {
            'method': 'GET',
            'url': '/config/query/{}',
            'decorators': CONFIG_ENTITY_DECORATORS,
            'description': 'Fetch the named <em>Query Configuration</em>'},
        'remove': {
            'method': 'DELETE',
            'url': '/config/query/{}',
            'decorators': CONFIG_ENTITY_DECORATORS,
            'description': 'Remove the named <em>Query Configuration</em>'},
        'replace': {
            'method': 'PUT',
            'url': '/config/query/{}',
            'decorators': CONFIG_ENTITY_DECORATORS,
            'description': 'Replace the named <em>Query Configuration</em>'},
        'data': {
            'method': 'GET',
            'url': '/data/query/{}',
            'decorators': DATA_RESPONSE_DECORATORS,
            'description': 'Execute the timeseries query specified by the named <em>Query Configuration</em>'},
        'execute': {
            'method': 'POST',
            'url': '/data/query',
            'decorators': DATA_RESPONSE_DECORATORS,
            'description': 'Execute the timeseries query specified in the request'},
    }



"""Client configuration"""
from typing import (
    Optional, Mapping, MutableMapping, Any
)
import os
import re
from pathlib import Path
from getpass import getpass
import json
import logging
import httpx
from appdirs import user_config_dir
from .auth import (
    WaylayCredentials,
    ClientCredentials, ApplicationCredentials, TokenCredentials, NoCredentials,
    WaylayTokenAuth, WaylayToken,
    parse_credentials,
    DEFAULT_ACCOUNTS_URL
)
from .exceptions import AuthError, ConfigError

log = logging.getLogger(__name__)

# http client dependencies
_http = httpx

TenantSettings = Mapping[str, str]
Settings = MutableMapping[str, str]

DEFAULT_PROFILE = '_default_'
SERVICE_KEY_API = 'waylay_api'


def tell(message: str):
    """interactive authentication message"""
    print(message)


def ask(prompt: str, secret: bool = False) -> str:
    """prompt user for credential information"""
    if secret:
        return getpass(prompt=prompt)
    else:
        return input(prompt)


class WaylayConfig():
    """Manages the authentication and endpoint configuration for the Waylay Platform"""
    profile: str
    settings_overrides: Settings
    credentials: WaylayCredentials
    _auth: _http.Auth
    _resolved_settings: Optional[TenantSettings] = None

    def __init__(
        self, credentials: WaylayCredentials = None, profile: str = DEFAULT_PROFILE,
        settings: TenantSettings = None, fetch_tenant_settings=True,
        auth: WaylayTokenAuth = None
    ):
        self.profile = profile
        if credentials is None:
            credentials = NoCredentials()
        self.credentials = credentials
        self.settings_overrides = dict(settings) if settings is not None else {}
        if auth is None:
            auth = WaylayTokenAuth(credentials, interactive=self.interactive)
        self._auth = auth
        if not fetch_tenant_settings:
            self._resolved_settings = {}

    def get_root_url(
        self, config_key: str, default_root_url: str = None, local=False
    ) -> Optional[str]:
        """get the root url for a waylay service"""
        config_key = _root_url_key_for(config_key)
        if default_root_url is None and config_key == SERVICE_KEY_API:
            default_root_url = _root_url_for(self.get_valid_token().domain)
        settings = self._local_settings if local else self.settings
        url = settings.get(config_key, default_root_url)
        if url is not None:
            return _root_url_for(url)
        return None

    def set_root_url(self, config_key: str, root_url: Optional[str]):
        """
        Overrides the root url for the given server (will persist on 'save').
        Setting a None value will remove the override.
        """
        config_key = _root_url_key_for(config_key)
        if root_url is None and config_key in self.settings_overrides:
            del self.settings_overrides[config_key]
        if root_url is not None:
            self.settings_overrides[config_key] = root_url

    @property
    def accounts_url(self):
        """accounts url (normalised)"""
        return _root_url_for(self.credentials.accounts_url)

    def interactive(self) -> WaylayCredentials:
        """ask interactively for credentials"""

        tell("Authenticating to the Waylay Platform")

        accounts_url = self.accounts_url
        acc_validated = False
        while not acc_validated:
            tell(f'Using authentication provider at [{accounts_url}]')
            accounts_url = ask(
                '> alternative endpoint (press enter to continue)?: '
            ) or accounts_url
            accounts_url = _root_url_for(accounts_url)
            accounts_status_resp = httpx.get(accounts_url)
            acc_validated = not accounts_status_resp.is_error
            if acc_validated:
                tell(f"Authenticating at '{accounts_status_resp.json()['name']}'")
            else:
                tell(f"Cannot connect to '{accounts_url}': {accounts_status_resp.reason_phrase}")

        tell("Please provide client credentials for the waylay data client.")
        credentials = ClientCredentials(api_key='', api_secret='', accounts_url=accounts_url)
        retry = 0
        while not credentials.is_well_formed() and retry < 3:
            api_key = ask(prompt='> apiKey : ', secret=False).strip()
            api_secret = ask(prompt='> apiSecret : ', secret=True).strip()
            credentials = ClientCredentials(
                api_key=api_key, api_secret=api_secret, accounts_url=accounts_url
            )
            if not credentials.is_well_formed():
                retry += 1
                if retry >= 3:
                    tell('Too many attempts, failing authentication')
                    raise AuthError('Too many attempts, failing authentication')
                tell('Invalid apiKey or apiSecret, please retry')
        self.credentials = credentials
        store = ''
        while not store or store[0].lower() not in ('n', 'y', 't', 'f'):
            store = ask(
                f'> Do you want to store these credentials (profile={self.profile})? [Y]: '
            ) or 'Y'
        if store[0].lower() in ('y', 't'):
            self.save()
            tell(
                f"Credential configuration stored as \n\t{self.config_file_path(self.profile)}\n"
                "Please make sure this file is treated securely.\n"
                "If compromised, _Revoke_ the api-key on the Waylay Console!"
            )
        return self.credentials

    @property
    def tenant_settings(self) -> TenantSettings:
        """
        Tenant settings as stored on accounts.
        Will fetch settings when not present and initialised with 'fetch_tenant_settings=True'.
        """
        if self._resolved_settings is None:
            self._resolved_settings = self._request_settings()

        return self._resolved_settings

    @property
    def settings(self) -> TenantSettings:
        """
        The tenant settings, as resolved form the accounts backend,
        and overridden with client config settings
        """
        return {
            **self.tenant_settings,
            **self.settings_overrides
        }

    @property
    def _local_settings(self) -> TenantSettings:
        """
        The locally available settings, including resolved settings
        from the backend if already present.
        """
        return {
            **(self._resolved_settings or {}),
            **self.settings_overrides
        }

    def get_valid_token(self) -> WaylayToken:
        """get the current valid authentication token or fail"""
        if isinstance(self.auth, WaylayTokenAuth):
            try:
                return self.auth.assure_valid_token()
            except AuthError as exc:
                raise ConfigError(f'cannot get valid token') from exc
        raise ConfigError('not using token authentication')

    @property
    def auth(self) -> _http.Auth:
        """returns the http authentication interceptor"""
        return self._auth

    @property
    def domain(self) -> Optional[str]:
        """the domain of the current user"""
        try:
            return self.get_valid_token().domain
        except ConfigError:
            return None

    def _request_settings(self) -> TenantSettings:
        try:
            settings_url = f"{self.get_root_url('api', local=True)}/api/settings"
            settings_resp = _http.get(settings_url, auth=self.auth)
            settings_resp.raise_for_status()
            return {
                key: value
                for key, value in settings_resp.json().items()
                if key.startswith('waylay_')
            }
        except _http.HTTPStatusError as exc:
            if exc.response.status_code == 403:
                log.warning(
                    "You are not authorised to fetch tenant settings.\n"
                    "The Waylay SAAS defaults will be used, unless you\n"
                    "provide explicit overrides in the SDK Configuration profile."
                )
                return {}
            raise ConfigError('cannot resolve tenant settings') from exc
        except _http.HTTPError as exc:
            raise ConfigError('cannot resolve tenant settings') from exc

    # config persistency methods
    @classmethod
    def config_file_path(cls, profile: str = DEFAULT_PROFILE) -> str:
        return os.path.join(user_config_dir('Waylay'), 'python_sdk', f".profile.{profile}.json")

    @classmethod
    def load(
        cls,
        profile: str = DEFAULT_PROFILE,
        interactive: bool = True,
        accounts_url: str = DEFAULT_ACCOUNTS_URL
    ):
        """load a stored waylay configuration"""
        try:
            with open(cls.config_file_path(profile), mode='r') as config_file:
                config_json = json.load(config_file)
            return cls.from_dict(config_json)
        except FileNotFoundError as exc:
            if not interactive:
                raise ConfigError(f'Config profile {profile} not found') from exc
            instance = cls(NoCredentials(accounts_url=accounts_url), profile=profile)
            return instance

    @classmethod
    def from_dict(cls, config_json: Mapping[str, Any]):
        config_json = dict(config_json)
        if 'credentials' in config_json:
            config_json['credentials'] = parse_credentials(config_json['credentials'])
        return cls(**config_json)

    def to_dict(self, obfuscate=True):
        return {
            'credentials': self.credentials.to_dict(obfuscate),
            'profile': self.profile,
            'settings': self.settings_overrides
        }

    def save(self):
        config_path = Path(self.config_file_path(self.profile))
        config_path.parent.mkdir(parents=True, exist_ok=True)
        with open(config_path, mode='w') as config_file:
            json.dump(self.to_dict(obfuscate=False), config_file)
            log.info("wrote waylay configuration: %s", config_path)

    @classmethod
    def delete(cls, profile: str = DEFAULT_PROFILE):
        """delete a stored profile"""
        config_path = Path(cls.config_file_path(profile))
        if config_path.exists():
            config_path.unlink()
            log.warning("waylay configuration removed: %s", config_path)
        else:
            log.warning("waylay configuration not found: %s", config_path)

    @classmethod
    def list_profiles(cls):
        """list stored profiles"""
        config_dir = Path(cls.config_file_path()).parent
        return {
            re.match(r'\.profile\.(.*)\.json', config_file.name)[1]: str(config_file)
            for config_file in config_dir.iterdir()
        }

    def __repr__(self):
        return f'<WaylayConfig({str(self)})>'

    def __str__(self):
        return json.dumps(self.to_dict(obfuscate=True))


def _root_url_key_for(config_key: Optional[str] = None):
    if config_key is None:
        return SERVICE_KEY_API
    if config_key.startswith('waylay_'):
        return config_key
    return f"waylay_{config_key}"


def _root_url_for(config_value: Optional[str]) -> Optional[str]:
    """make sure any host name is converted in a https:// url"""
    if config_value and '://' not in config_value:
        return f'https://{config_value}'
    return config_value



"""Resource action method decorators that are generally usefull"""

from functools import wraps
from typing import (
    Union, List, Mapping, Any, Callable
)
from simple_rest_client.exceptions import ErrorWithResponse
from ..exceptions import RestResponseError, RestResponseParseError


def suppress_header_decorator(header_key):
    """decorator that suppresses a configured header on a resource during execution"""
    def decorator(action_method):
        @wraps(action_method)
        def wrapped(slf, *args, **kwargs):
            header_value = slf.headers.pop(header_key, None)
            try:
                return action_method(slf, *args, **kwargs)
            finally:
                if header_value:
                    slf.headers[header_key] = header_value
        return wrapped
    return decorator


def exception_decorator(action_method):
    """
    decorator that parses json error responses
    """
    @wraps(action_method)
    def wrapped(*args, **kwargs):
        try:
            return action_method(*args, **kwargs)
        except ErrorWithResponse as exc:
            raise RestResponseError.from_cause(exc) from exc
    return wrapped


def identity_transform(x):
    return x


def return_path_decorator(
    default_path: List[Union[int, str]],
    default_response_constructor: Callable[[Any], Any] = identity_transform
):
    """
    Decorator method that extracts a part of the json body of an action response.
    The 'default_path' is a list of 'int' or 'str' keys that select through
    'dict' and 'list' structures.
    The option 'response_constructor' wrap the value in a class or constructor method
    It processes the following qualified arguments in a method call:
    * if 'raw=True' is set, the original response is returned
    * if 'select_path' is set, that path is used rather than the 'default_path'
       provided in the decorator constructor.
    * if 'response_constructor' is set, it is used to wrap the response
      rather than 'default_response_constructor'
    """
    def decorator(action_method):
        @wraps(action_method)
        def wrapped(*args, **kwargs):
            raw = kwargs.pop('raw', False)
            select_path = kwargs.pop('select_path', default_path)
            constructor = kwargs.pop(
                'response_constructor', default_response_constructor
            ) or identity_transform
            response = action_method(*args, **kwargs)
            if raw:
                return response
            try:
                return constructor(
                    _select_path_from(response.body, [], select_path)
                )
            except AttributeError as exc:
                raise RestResponseParseError(exc.args[0], response) from exc
        return wrapped
    return decorator


def return_body_decorator(action_method):
    """
    Decorator that returns the body of an action response, unless a 'raw=True' is
    provided in the call
    """
    return return_path_decorator([], identity_transform)(action_method)


def _select_path_from(
    value: Any, ctx_path: List[Union[int, str]], path: List[Union[int, str]]
) -> Any:
    """
    Recursively select values from 'Mapping' (dict, 'str' key) input,
    or positional elements from 'List' (list, 'int' key) input,
    with a 'path' consisting of a list of keys.
    Applied keys are moved to the 'ctx_path' to be used for error reporting.
    'str' keys applied to a 'List' are mapped through to the underlying structure.
    """
    if not path:
        return value
    key = path[0]
    new_ctx_path = ctx_path + [key]
    rem_path = path[1:]
    if isinstance(value, List):
        if isinstance(key, int) and len(value) < key:
            return _select_path_from(value[key], new_ctx_path, rem_path)
        if isinstance(key, str):
            return [
                _select_path_from(element[key], new_ctx_path, rem_path)
                for element in value
            ]
    if isinstance(value, Mapping) and isinstance(key, str) and key in value:
        return _select_path_from(value[key], new_ctx_path, rem_path)

    ctx_path_str = '.'.join(str(p) for p in new_ctx_path)
    raise AttributeError(f"Cannot find '{ctx_path_str}' in response body.")



"""waylay-beta build configuration"""
from setuptools import setup, find_namespace_packages
import versioneer

setup(
    name='waylay-beta',
    description='beta release of the Waylay Python SDK',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    url='https://docs.waylay.io/api/sdk/python',
    author='Waylay',
    author_email='info@waylay.io',
    license='ISC',
    license_file='LICENSE.txt',
    packages=find_namespace_packages(),
    package_data={"waylay": ["py.typed"]},
    include_package_data=True,
    install_requires=[
        'httpx',
        'simple-rest-client',
        'appdirs',
        'python-jose',
        'pandas',
        'isodate',
        'joblib'
    ],
    extras_require={
        'dev': [
            'pytest',
            'pytest-cov',
            'pytest-mock',
            'mock',
            'pylint',
            'pycodestyle',
            'autopep8',
            'mypy',
            'typing-inspect',
            'sklearn'
        ],
        ':python_version == "3.6"': [
            'dataclasses',
            'typing_extensions'
        ],
        ':python_version == "3.7"': [
            'typing_extensions'
        ],
        'ml': [
            'sklearn',
            'tensorflow'
        ]
    },
    setup_requires=[
        'setuptools-pep8'
    ],
    python_requires='>=3.6',
    entry_points={
        "console_scripts": [
            "waylaycli = waylay.cli.waylaycli:main"
        ],
        "waylay_services": [
            "beta = waylay.service:SERVICES"
        ]
    }
)
